/* HeuristicLab
* Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
*
* This file is part of HeuristicLab.
*
* HeuristicLab is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* HeuristicLab is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
*
*
* Export qhull functionality for HeuristicLab. 
* Code based on qhull's user_eg.c as well as
* R's geometry package (Rconvhulln.c, Copyright Kai Habel, Raoul Grasman, David Sterratt; licensed under the GNU GPL).
* 
*/     

#define qh_QHimport
#include "qhull_a.h"

#ifdef __cplusplus
extern "C" {
#endif
  __declspec(dllexport) double qhull_volume(int dim, int numpoints, double *data);
  __declspec(dllexport) int* qhull_convex_hull(int dim, int numpoints, double *data, int *retCode, int* nrOfFacets); 
  __declspec(dllexport) void qhull_free(int* data);
#ifdef __cplusplus
}
#endif

void print_summary (void) {
  facetT *facet;
  int k;

  printf ("\n%d vertices and %d facets with normals:\n",
    qh num_vertices, qh num_facets);
  FORALLfacets {
    for (k=0; k < qh hull_dim; k++)
      printf ("%6.2g ", facet->normal[k]);
    printf ("\n");
  }
}          

double qhull_volume(int dim, int numpoints, double *data) {
  boolT ismalloc= False;   
  char flags[250];       
  FILE *outfile= NULL;   
  FILE *errfile= stderr;   
  facetT *facet;           
  int curlong, totlong;    
  int i,j, exitcode;
  double volume = -1.0;

#if qh_QHpointer  
  if (qh_qh){
    printf ("QH6233: Qhull link error.  The global variable qh_qh was not initialized\n\
            to NULL by global.c.  Please compile user_eg.c with -Dqh_QHpointer_dllimport\n\
            as well as -Dqh_QHpointer, or use libqhullstatic, or use a different tool chain.\n\n");
    return -1;
  }
#endif

  sprintf (flags, "qhull s Tv Qt FA"); 

  exitcode = qh_new_qhull (dim, numpoints, data, ismalloc, flags, outfile, errfile);
  if (!exitcode) {                      
   // print_summary();  
    qh_vertexneighbors();
    volume = qh totvol;
  } else {
    return -1.0;
  }
  qh_freeqhull(!qh_ALL);            
  qh_memfreeshort (&curlong, &totlong); 
  if (curlong || totlong)
    fprintf (errfile, "qhull internal warning (HeuristicLab_qhull, #1): did not free %d bytes of long memory (%d pieces)\n", totlong, curlong);

  return volume;
}


int* qhull_convex_hull(int dim, int numpoints, double *data, int *retCode, int* nrOfFacets) {
  boolT ismalloc= False;   
  char flags[250];       
  FILE *outfile= NULL;   
  FILE *errfile= stderr;   
  facetT *facet;           
  int curlong, totlong;    
  int i,j, exitcode;
  int* result = NULL; 

#if qh_QHpointer  
  if (qh_qh){
    printf ("QH6233: Qhull link error.  The global variable qh_qh was not initialized\n\
            to NULL by global.c.  Please compile user_eg.c with -Dqh_QHpointer_dllimport\n\
            as well as -Dqh_QHpointer, or use libqhullstatic, or use a different tool chain.\n\n");
    *retCode = -1;
    return NULL;
  }
#endif

  sprintf (flags, "qhull s Tv Qt FA");

  exitcode= qh_new_qhull (dim, numpoints, data, ismalloc, flags, outfile, errfile);
  if (!exitcode) {      
    facetT *facet; 
    vertexT *vertex, **vertexp;
    unsigned int n = qh num_facets;
    *nrOfFacets = n;

    //print_summary(); 
    qh_vertexneighbors();

    result = (int *) malloc(n*dim*sizeof(int));     

    i=0;
    FORALLfacets {
      j=0;      
      FOREACHvertex_ (facet->vertices) {        
        if (j >= dim)
          fprintf (errfile,"extra vertex %d of facet %d = %d",j++,i,/*1+*/qh_pointid(vertex->point));    //TODO: handle in return value
        else
          result[i+n*j++] = qh_pointid(vertex->point);
      }
      if (j < dim) fprintf (errfile,"facet %d only has %d vertices",i,j);
      i++;
    }
  } else { 
    *retCode = -1;
    return NULL;
  }
  qh_freeqhull(!qh_ALL);            
  qh_memfreeshort (&curlong, &totlong); 
  if (curlong || totlong)
    fprintf (errfile, "qhull internal warning (HeuristicLab_qhull, #1): did not free %d bytes of long memory (%d pieces)\n", totlong, curlong);

  return result;
}


void qhull_free(int* data) {
  free(data);
}


